print("import: 'paste'")
import paste

print("import: 'paste.auth'")
import paste.auth

print("import: 'paste.cowbell'")
import paste.cowbell

print("import: 'paste.debug'")
import paste.debug

print("import: 'paste.evalexception'")
import paste.evalexception

print("import: 'paste.exceptions'")
import paste.exceptions

print("import: 'paste.util'")
import paste.util

